'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const unplugin$1 = require('unplugin');
const node_module = require('node:module');
const path = require('node:path');
const typescript = require('@volar/typescript');
const vue = require('@vue/language-core');
const MagicString = require('magic-string');
const ts = require('typescript');

var _documentCurrentScript = typeof document !== 'undefined' ? document.currentScript : null;
function _interopDefaultCompat (e) { return e && typeof e === 'object' && 'default' in e ? e.default : e; }

function _interopNamespaceCompat(e) {
	if (e && typeof e === 'object' && 'default' in e) return e;
	const n = Object.create(null);
	if (e) {
		for (const k in e) {
			n[k] = e[k];
		}
	}
	n.default = e;
	return n;
}

const path__namespace = /*#__PURE__*/_interopNamespaceCompat(path);
const vue__namespace = /*#__PURE__*/_interopNamespaceCompat(vue);
const MagicString__default = /*#__PURE__*/_interopDefaultCompat(MagicString);
const ts__default = /*#__PURE__*/_interopDefaultCompat(ts);

const windowsPathReg = /\\/g;
const normalizePath = (id) => id.replace(windowsPathReg, "/");
const resolveOptions = (rawOptions) => ({
  tsconfigPath: rawOptions.tsconfigPath ?? path.join(process.cwd(), "tsconfig.json"),
  defineEmits: rawOptions.defineEmits ?? true,
  defineProps: rawOptions.defineProps ?? true
});
const quotesReg = /"/g;
const escapeQuotes = (s) => s.replace(quotesReg, '\\"');

const require$1 = node_module.createRequire((typeof document === 'undefined' ? require('u' + 'rl').pathToFileURL(__filename).href : (_documentCurrentScript && _documentCurrentScript.src || new URL('index.cjs', document.baseURI).href)));
function createLanguage(tsconfigPath, ts = require$1("typescript")) {
  const tsconfig = normalizePath(tsconfigPath);
  return createLanguageWorker(
    () => vue__namespace.createParsedCommandLine(ts, ts.sys, tsconfigPath),
    path__namespace.dirname(tsconfig),
    ts
  );
}
function createLanguageWorker(loadParsedCommandLine, rootPath, ts) {
  let parsedCommandLine = loadParsedCommandLine();
  let fileNames = parsedCommandLine.fileNames.map(normalizePath);
  let projectVersion = 0;
  const scriptSnapshots = /* @__PURE__ */ new Map();
  const _host = {
    workspacePath: rootPath,
    rootPath,
    getProjectVersion: () => projectVersion.toString(),
    getCompilationSettings: () => parsedCommandLine.options,
    getScriptFileNames: () => fileNames,
    getProjectReferences: () => parsedCommandLine.projectReferences,
    getScriptSnapshot: (fileName) => {
      if (!scriptSnapshots.has(fileName)) {
        const fileText = ts.sys.readFile(fileName);
        if (fileText !== void 0) {
          scriptSnapshots.set(fileName, ts.ScriptSnapshot.fromString(fileText));
        }
      }
      return scriptSnapshots.get(fileName);
    }
  };
  return {
    ...baseCreateLanguageWorker(
      _host,
      vue__namespace.resolveVueCompilerOptions(parsedCommandLine.vueOptions),
      ts
    ),
    updateFile(fileName, text) {
      fileName = normalizePath(fileName);
      scriptSnapshots.set(fileName, ts.ScriptSnapshot.fromString(text));
      projectVersion++;
    },
    deleteFile(fileName) {
      fileName = normalizePath(fileName);
      fileNames = fileNames.filter((f) => f !== fileName);
      projectVersion++;
    },
    reload() {
      parsedCommandLine = loadParsedCommandLine();
      fileNames = parsedCommandLine.fileNames.map(normalizePath);
      this.clearCache();
    },
    clearCache() {
      scriptSnapshots.clear();
      projectVersion++;
    }
  };
}
function baseCreateLanguageWorker(host, vueCompilerOptions, ts) {
  const vueLanguages = ts ? (
    // eslint-disable-next-line etc/no-deprecated
    vue__namespace.createLanguages(host.getCompilationSettings(), vueCompilerOptions, ts)
  ) : [];
  const core = vue__namespace.createLanguageContext(host, vueLanguages);
  const tsLsHost = typescript.createLanguageServiceHost(core, ts, ts.sys);
  const tsLs = ts.createLanguageService(tsLsHost);
  typescript.decorateLanguageService(core.virtualFiles, tsLs, false);
  const getScriptKind = tsLsHost.getScriptKind;
  tsLsHost.getScriptKind = (fileName) => {
    if (fileName.endsWith(".vue.js")) {
      return ts.ScriptKind.TS;
    }
    if (fileName.endsWith(".vue.jsx")) {
      return ts.ScriptKind.TSX;
    }
    return getScriptKind(fileName);
  };
  const program = tsLs.getProgram();
  const typeChecker = program.getTypeChecker();
  function findNode(filepath, filter) {
    filepath = normalizePath(filepath);
    const sourceFile = core.virtualFiles.getSource(filepath)?.root;
    if (!(sourceFile instanceof vue__namespace.VueFile)) {
      return;
    }
    if (!sourceFile.sfc.scriptSetup) {
      return;
    }
    const {
      ast: scriptSetupAst,
      startTagEnd: offset,
      lang
    } = sourceFile.sfc.scriptSetup;
    if (lang !== "ts" && lang !== "tsx") {
      return;
    }
    const virtualFileAst = program.getSourceFile(`${filepath}.${lang}`);
    if (!virtualFileAst) {
      return;
    }
    const scriptSetupRanges = vue__namespace.parseScriptSetupRanges(
      ts,
      scriptSetupAst,
      vueCompilerOptions
    );
    const virtualTsFileScriptSetupRanges = vue__namespace.parseScriptSetupRanges(
      ts,
      virtualFileAst,
      vueCompilerOptions
    );
    const sourceRange = filter(scriptSetupRanges);
    const virtualTsRange = filter(virtualTsFileScriptSetupRanges);
    if (!sourceRange || !virtualTsRange) {
      return;
    }
    let foundVirtualFileNode;
    virtualFileAst.forEachChild(function traverse(node) {
      if (node.getStart(virtualFileAst) === virtualTsRange.start && node.getEnd() === virtualTsRange.end) {
        foundVirtualFileNode = node;
      } else {
        node.forEachChild(traverse);
      }
    });
    let foundSetupNode;
    scriptSetupAst.forEachChild(function traverse(node) {
      if (node.getStart(scriptSetupAst) === sourceRange.start && node.getEnd() === sourceRange.end) {
        foundSetupNode = node;
      } else {
        node.forEachChild(traverse);
      }
    });
    const setupRange = {
      start: offset + sourceRange.start,
      end: offset + sourceRange.end
    };
    return {
      virtualFileNode: foundVirtualFileNode,
      scriptNode: foundSetupNode,
      setupRange,
      scriptSetupAst,
      virtualFileAst,
      offset
    };
  }
  return {
    findNode,
    __internal__: {
      tsLs,
      program,
      typeChecker
    }
  };
}
const LANGUAGE_GLOBAL_KEY = "__UNPLUGIN_VUE_COMPLEX_LANGUAGE__";
function ensureLanguage(tsconfigPath) {
  if (!globalThis[LANGUAGE_GLOBAL_KEY]) {
    globalThis[LANGUAGE_GLOBAL_KEY] = createLanguage(tsconfigPath);
  }
}
function getLanguage() {
  if (!globalThis[LANGUAGE_GLOBAL_KEY]) {
    throw new Error("[unplugin-vue-complex-types] Language not created!");
  }
  return globalThis[LANGUAGE_GLOBAL_KEY];
}

class Printer {
  constructor(checker) {
    this.checker = checker;
  }
  getBaseType(nodeOrType) {
    if ("kind" in nodeOrType) {
      nodeOrType = this.checker.getTypeAtLocation(nodeOrType);
    }
    return this.checker.getBaseConstraintOfType(nodeOrType) ?? this.checker.getBaseTypeOfLiteralType(nodeOrType);
  }
  printIntersectionTypeNode(node) {
    return node.types.map((t) => this.printTypeArg(t)).join(" & ");
  }
  printUnionTypeNode(node) {
    return node.types.map((t) => this.printTypeArg(t)).join(" | ");
  }
  printTypeLiteralNode(node) {
    const parts = ["{"];
    for (const member of node.members) {
      if (ts__default.isPropertySignature(member)) {
        const stringBaseType = member.type ? this.checker.typeToString(this.getBaseType(member.type)) : "any";
        parts.push(
          [
            member.name.getText(),
            member.questionToken?.getText(),
            ": ",
            stringBaseType
          ].filter(Boolean).join("")
        );
      }
    }
    parts.push("}");
    return parts.join("\n");
  }
  printByType(node) {
    const type = this.checker.getTypeAtLocation(node);
    const properties = type.getProperties();
    const parts = ["{"];
    let questionToken = "";
    if (ts__default.isMappedTypeNode(node)) {
      questionToken = node.questionToken?.getText() ?? "";
    }
    for (const property of properties) {
      const valueType = this.checker.getTypeOfSymbol(property);
      const stringValueType = this.checker.typeToString(
        this.getBaseType(valueType)
      );
      parts.push(
        `${this.checker.symbolToString(
          property
        )}${questionToken}: ${stringValueType}`
      );
    }
    parts.push("}");
    return parts.join("\n");
  }
  printTypeArg(node) {
    if (ts__default.isIntersectionTypeNode(node)) {
      return this.printIntersectionTypeNode(node);
    } else if (ts__default.isUnionTypeNode(node)) {
      return this.printUnionTypeNode(node);
    } else if (ts__default.isTypeLiteralNode(node)) {
      return this.printTypeLiteralNode(node);
    } else if (ts__default.isMappedTypeNode(node) || ts__default.isTypeReferenceNode(node)) {
      return this.printByType(node);
    } else {
      console.error(
        `[unplugin-vue-complex-types] \`${ts__default.SyntaxKind[node.kind]}\` is not supported.`
      );
      return node.getText();
    }
  }
  printEventsByCallSignatures(callSignatures) {
    return callSignatures.map((c) => {
      const parameters = c.getParameters();
      const event = parameters[0];
      return this.checker.typeToString(this.checker.getTypeOfSymbol(event));
    });
  }
  printEventsByMembers(members) {
    return members.map((m) => `"${escapeQuotes(m.getName())}"`);
  }
  printEventsRuntimeArg(node) {
    const parts = [];
    const type = this.checker.getTypeAtLocation(node);
    const callSignatures = type.getCallSignatures();
    const members = type.getProperties();
    if (callSignatures.length > 0 && members.length > 0) {
      throw new Error(
        "[unplugin-vue-complex-types] You may not use old style `defineEmits` and `defineEmits` shorthand together."
      );
    }
    if (members.length > 0) {
      parts.push(...this.printEventsByMembers(members));
    } else {
      parts.push(...this.printEventsByCallSignatures(callSignatures));
    }
    return `[${parts.join(", ")}]`;
  }
}

function transformDefineProps(printer, s, id) {
  const language = getLanguage();
  const definePropsTypeArg = language.findNode(
    id,
    (scriptSetupRanges) => scriptSetupRanges.props.define?.typeArg
  );
  if (!definePropsTypeArg) {
    return;
  }
  const { virtualFileNode: typeArgNode, setupRange: typeArgRange } = definePropsTypeArg;
  const printedType = printer.printTypeArg(typeArgNode);
  s.overwrite(typeArgRange.start, typeArgRange.end, printedType);
}
function transformDefineEmits(printer, s, id) {
  const language = getLanguage();
  const defineEmits = language.findNode(
    id,
    (scriptSetupRanges) => scriptSetupRanges.emits.define
  );
  if (!defineEmits) {
    return;
  }
  const {
    scriptNode: defineEmitsNode,
    virtualFileNode: virtualFileDefineEmitsNode,
    scriptSetupAst,
    offset
  } = defineEmits;
  const defineEmitsTypeArg = virtualFileDefineEmitsNode && ts__default.isCallExpression(virtualFileDefineEmitsNode) && virtualFileDefineEmitsNode.typeArguments?.[0];
  if (!defineEmitsTypeArg) {
    return;
  }
  const tokens = defineEmitsNode.getChildren(scriptSetupAst);
  const lessThanToken = tokens.find(
    (t) => t.kind === ts__default.SyntaxKind.LessThanToken
  );
  const greaterThanToken = tokens.find(
    (t) => t.kind === ts__default.SyntaxKind.GreaterThanToken
  );
  const openParenToken = tokens.find(
    (t) => t.kind === ts__default.SyntaxKind.OpenParenToken
  );
  if (!lessThanToken || !greaterThanToken || !openParenToken) {
    return;
  }
  const defineEmitsTypeArgRange = [
    offset + lessThanToken.pos,
    offset + greaterThanToken.end
  ];
  const runtimeArgPos = offset + openParenToken.end;
  const printedRuntimeArg = printer.printEventsRuntimeArg(defineEmitsTypeArg);
  s.remove(...defineEmitsTypeArgRange);
  s.appendRight(runtimeArgPos, printedRuntimeArg);
}
function transform(code, id, options) {
  const s = new MagicString__default(code);
  const language = getLanguage();
  const typeChecker = language.__internal__.typeChecker;
  const printer = new Printer(typeChecker);
  if (options.defineProps) {
    transformDefineProps(printer, s, id);
  }
  if (options.defineEmits) {
    transformDefineEmits(printer, s, id);
  }
  return {
    code: s.toString(),
    map: s.generateMap({ hires: true })
  };
}

const unplugin = unplugin$1.createUnplugin((options = {}) => ({
  name: "unplugin-vue-complex-types",
  buildStart() {
    const resolvedOptions = resolveOptions(options);
    ensureLanguage(resolvedOptions.tsconfigPath);
  },
  transform(code, id) {
    const resolvedOptions = resolveOptions(options);
    if (!id.endsWith(".vue")) {
      return;
    }
    return transform(code, id, resolvedOptions);
  }
}));

exports.default = unplugin;
exports.ensureLanguage = ensureLanguage;
exports.getLanguage = getLanguage;
